from rest_framework import routers

from . import providers
from . import service_areas
from . import vehicles


def get_prv_router():
    """Generates a fresh router.

    Enables to register new routes even after .urls has
    been called on the router.
    """
    prv_router = routers.DefaultRouter()
    prv_router.register(r"providers", providers.ProviderViewSet)
    prv_router.register(r"service_areas", service_areas.AreaViewSet)
    prv_router.register(r"polygons", service_areas.PolygonViewSet)
    prv_router.register(r"vehicles", vehicles.DeviceViewSet, basename="device")
    return prv_router


app_name = "mds_prv_api"
urlpatterns = get_prv_router().urls



from rest_framework import serializers
from rest_framework import viewsets

from mds import models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_ADMIN


class ProviderSerializer(serializers.ModelSerializer):
    id = serializers.UUIDField(
        required=False, help_text="Unique provider identifier (UUID)"
    )
    name = serializers.CharField(help_text="Name of the Provider")
    logo_b64 = serializers.CharField(
        required=False, help_text="Logo of provider base64 encoded"
    )

    class Meta:
        model = models.Provider
        fields = ("id", "name", "logo_b64")


class ProviderViewSet(viewsets.ModelViewSet):
    permission_classes = (require_scopes(SCOPE_ADMIN),)
    queryset = models.Provider.objects.all()
    lookup_field = "id"
    serializer_class = ProviderSerializer



import json

from rest_framework import serializers
from rest_framework import viewsets

from mds import models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_VEHICLE


class MultiPolygonField(serializers.Field):
    def to_representation(self, value):
        ret = {"type": "MultiPolygon", "coordinates": []}
        for polygon in value:
            ret["coordinates"].append(json.loads(polygon.geom.geojson)["coordinates"])
        return ret

    def to_internal_value(self, value):
        raise NotImplementedError()


class AreaSerializer(serializers.ModelSerializer):
    """A service area
    """

    service_area_id = serializers.UUIDField(
        source="id", help_text="Unique Area identifier (UUID)"
    )
    service_area = MultiPolygonField(source="polygons.all")

    class Meta:
        model = models.Area
        fields = ("service_area_id", "service_area")


class AreaViewSet(viewsets.ReadOnlyModelViewSet):
    permission_classes = (require_scopes(SCOPE_VEHICLE),)
    queryset = models.Area.objects.prefetch_related("polygons").all()
    lookup_field = "id"
    serializer_class = AreaSerializer

    def get_queryset(self):
        queryset = super().get_queryset()
        provider_id = getattr(self.request.user, "provider_id", None)
        if provider_id:
            queryset = queryset.filter(providers__id=provider_id)
        else:
            queryset = queryset.none()

        return queryset



"""A few helpers around DRF."""
import datetime
import json

from django.contrib.gis.geos.geometry import GEOSGeometry
import drf_yasg.inspectors
import drf_yasg.inspectors.base
import drf_yasg.utils
from rest_framework import serializers
from rest_framework.response import Response


# Viewsets #####################################################


class MultiSerializerViewSetMixin:
    serializers_mapping = {}

    def _create(self, request, *args, **kwargs):
        """Patch of rest_framework.mixins.CreateModelMixin.create.

        This patch works with our Request/Response serializers.
        """
        request_serializer = self.get_serializer(
            data=request.data, context={"request_or_response": "request"}
        )
        request_serializer.is_valid(raise_exception=True)
        instance = request_serializer.save()
        response_serializer = self.get_serializer(
            instance=instance, context={"request_or_response": "response"}
        )
        return Response(response_serializer.data, status=201)

    def _update(self, request, *args, **kwargs):
        """Patch of rest_framework.mixins.UpdateModelMixin.create.

        This patch works with our Request/Response serializers.
        """
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        request_serializer = self.get_serializer(
            instance, data=request.data, partial=partial
        )
        request_serializer.is_valid(raise_exception=True)
        request_serializer.save()

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        response_serializer = self.get_serializer(
            instance=instance, context={"request_or_response": "response"}
        )
        return Response(response_serializer.data)

    def get_serializer(self, *args, **kwargs):
        context = kwargs.get("context", {})
        context.update(self.get_serializer_context())
        kwargs["context"] = context
        request_or_response = context.get("request_or_response", "response")
        serializer = self.serializers_mapping.get(self.action, {}).get(
            request_or_response
        )
        if serializer:
            return serializer(*args, **kwargs)
        return super().get_serializer(*args, **kwargs)


# Serializers ##################################################


class EmptyResponseSerializer(serializers.Serializer):
    pass


class GeometryField(serializers.Field):
    type_name = "GeometryField"

    def to_representation(self, value):
        if isinstance(value, dict) or value is None:
            return value
        return json.loads(value.geojson)

    def to_internal_value(self, value):
        if isinstance(value, GEOSGeometry) or value is None:
            return value
        if isinstance(value, dict):
            value = json.dumps(value)
        return GEOSGeometry(value)


class UnixTimestampMilliseconds(serializers.IntegerField):
    def to_representation(self, value: datetime.datetime):
        dt = value - datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
        return int(dt.total_seconds() * 1000)

    def to_internal_value(self, value):
        value = super().to_representation(value)
        dt = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
        return dt + datetime.timedelta(0, 0, 0, value)


# Schema #########################################################


def _call_view_method(
    view, method_name, fallback_attr=None, default=None, args=None, kwargs=None
):
    """Override of drf_yasg.inspectors.base.call_view_method to allow passing args."""
    if hasattr(view, method_name):
        try:
            view_method, is_callabale = drf_yasg.inspectors.base.is_callable_method(
                view, method_name
            )
            if is_callabale:
                args = args or []
                kwargs = kwargs or {}
                return view_method(*args, **kwargs)
        except Exception:  # noqa
            drf_yasg.inspectors.base.logger.warning(
                "view's %s raised exception during schema generation; use "
                "`getattr(self, 'swagger_fake_view', False)` to detect and short-circuit this",
                type(view).__name__,
                exc_info=True,
            )

    if fallback_attr and hasattr(view, fallback_attr):
        return getattr(view, fallback_attr)

    return default


class CustomSwaggerAutoSchema(drf_yasg.inspectors.SwaggerAutoSchema):
    """Customer Schema generator that understands Request and Response serializers.

    By default DRF has no concept of Request serializer VS Response serializer.
    A view only defines one serializer which could be limiting.
    This Schema generator, in combination with MultiSerializerViewSetMixin bypasses
    this limitation.
    """

    def get_view_serializer(self, request_or_response):
        """Return the serializer as defined by the view's ``get_serializer()`` method.

        :return: the view's ``Serializer``
        :rtype: rest_framework.serializers.Serializer
        """
        return _call_view_method(
            self.view,
            "get_serializer",
            kwargs={"context": {"request_or_response": request_or_response}},
        )

    def get_request_serializer(self):
        """Return the request serializer (used for parsing the request payload) for this endpoint.

        :return: the request serializer, or one of :class:`.Schema`, :class:`.SchemaRef`, ``None``
        :rtype: rest_framework.serializers.Serializer
        """
        body_override = self._get_request_body_override()

        if body_override is None and self.method in self.implicit_body_methods:
            return self.get_view_serializer("request")

        if body_override is drf_yasg.utils.no_body:
            return None

        return body_override

    def get_default_response_serializer(self):
        """Return the default response serializer for this endpoint. This is derived from either the ``request_body``
        override or the request serializer (:meth:`.get_view_serializer`).

        :return: response serializer, :class:`.Schema`, :class:`.SchemaRef`, ``None``
        """
        body_override = self._get_request_body_override()
        if body_override and body_override is not drf_yasg.utils.no_body:
            return body_override

        return self.get_view_serializer("response")



from rest_framework import serializers
from rest_framework import viewsets

from mds import models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_ADMIN
from mds.apis import utils


class PolygonSerializer(serializers.ModelSerializer):
    """A geographic polygon.
    """

    id = serializers.UUIDField(
        required=False, help_text="Unique Polygon identifier (UUID)"
    )
    label = serializers.CharField(help_text="Name of the polygon")
    creation_date = serializers.DateTimeField(help_text="Polygon creation date")
    deletion_date = serializers.DateTimeField(
        required=False, help_text="Polygon deletion date"
    )
    properties = serializers.JSONField(help_text="Properties of the Polygon")
    geom = utils.GeometryField(help_text="GeoJSON Polygon")

    class Meta:
        model = models.Polygon
        fields = ("id", "label", "creation_date", "deletion_date", "geom", "properties")


class AreaSerializer(serializers.ModelSerializer):
    """A service area, composed of a group of Polygons.
    """

    id = serializers.UUIDField(
        required=False, help_text="Unique Area identifier (UUID)"
    )
    label = serializers.CharField(help_text="Name of the Area")
    creation_date = serializers.DateTimeField(help_text="Area creation date")
    deletion_date = serializers.DateTimeField(
        required=False, help_text="Area deletion date"
    )
    polygons = PolygonSerializer(required=False, many=True)

    class Meta:
        model = models.Area
        fields = ("id", "label", "creation_date", "deletion_date", "polygons")


class PolygonViewSet(viewsets.ModelViewSet):
    permission_classes = (require_scopes(SCOPE_ADMIN),)
    queryset = models.Polygon.objects.all()
    lookup_field = "id"
    serializer_class = PolygonSerializer


class AreaViewSet(viewsets.ModelViewSet):
    permission_classes = (require_scopes(SCOPE_ADMIN),)
    queryset = models.Area.objects.prefetch_related("polygons").all()
    lookup_field = "id"
    serializer_class = AreaSerializer



# Generated by Django 2.1.4 on 2019-01-17 00:08

import django.contrib.gis.db.models.fields
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import mds.models
import rest_framework.utils.encoders
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Area',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('deletion_date', models.DateTimeField(null=True)),
                ('label', mds.models.UnboundedCharField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('registration_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('identification_number', mds.models.UnboundedCharField()),
                ('category', mds.models.UnboundedCharField(choices=[('bike', 'Bike'), ('scooter', 'Scooter'), ('car', 'Car')])),
                ('model', mds.models.UnboundedCharField(default=str)),
                ('propulsion', django.contrib.postgres.fields.ArrayField(base_field=mds.models.UnboundedCharField(choices=[('electric', 'Electric'), ('combustion', 'Combustion')]), size=None)),
                ('year_manufactured', models.IntegerField(null=True)),
                ('manufacturer', mds.models.UnboundedCharField(default=str)),
                ('dn_gps_point', django.contrib.gis.db.models.fields.PointField(null=True, srid=4326)),
                ('dn_gps_timestamp', models.DateTimeField(null=True)),
                ('dn_status', mds.models.UnboundedCharField(choices=[('available', 'Available'), ('reserved', 'Reserved'), ('unavailable', 'Unavailable'), ('removed', 'Removed')], default='unavailable')),
            ],
        ),
        migrations.CreateModel(
            name='EventRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('saved_at', models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
                ('source', models.CharField(choices=[('push', 'push'), ('pull', 'pull')], default='push', max_length=16)),
                ('event_type', mds.models.UnboundedCharField(choices=[('service_start', 'Service start'), ('trip_end', 'Trip end'), ('rebalance_drop_off', 'Rebalance drop_off'), ('maintenance_drop_off', 'Maintenance drop off'), ('cancel_reservation', 'Cancel reservation'), ('reserve', 'Reserve'), ('trip start', 'Trip start'), ('trip enter', 'Trip enter'), ('trip_leave', 'Trip leave'), ('register', 'Register'), ('low_battery', 'Low battery'), ('maintenance', 'Maintenance'), ('service_end', 'Service end'), ('rebalance_pick_up', 'Rebalance pick up'), ('maintenance_pick_up', 'Maintenance pick up'), ('deregister', 'Deregister'), ('telemetry', 'Received telemetry')])),
                ('properties', django.contrib.postgres.fields.jsonb.JSONField(default=dict, encoder=rest_framework.utils.encoders.JSONEncoder)),
                ('device', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='event_records', to='mds.Device')),
            ],
        ),
        migrations.CreateModel(
            name='Polygon',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('deletion_date', models.DateTimeField(null=True)),
                ('label', mds.models.UnboundedCharField(null=True)),
                ('geom', django.contrib.gis.db.models.fields.PolygonField(srid=4326)),
                ('properties', django.contrib.postgres.fields.jsonb.JSONField(default=dict, encoder=rest_framework.utils.encoders.JSONEncoder)),
            ],
        ),
        migrations.CreateModel(
            name='Provider',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('name', mds.models.UnboundedCharField(default=str)),
                ('logo_b64', mds.models.UnboundedCharField(blank=True, default=None, null=True)),
            ],
        ),
        migrations.AddField(
            model_name='device',
            name='provider',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='devices', to='mds.Provider'),
        ),
        migrations.AddField(
            model_name='area',
            name='polygons',
            field=models.ManyToManyField(blank=True, related_name='areas', to='mds.Polygon'),
        ),
        migrations.AddField(
            model_name='area',
            name='providers',
            field=models.ManyToManyField(blank=True, related_name='areas', to='mds.Provider'),
        ),
    ]



from django.utils.translation import ugettext_lazy as _


HTTP_METHOD_CHOICES = [("POST", "POST"), ("PUT", "PUT"), ("DELETE", "DELETE")]
DEVICE_STATUS_CHOICES = [
    ("available", _("Available")),
    ("reserved", _("Reserved")),
    ("unavailable", _("Unavailable")),
    ("removed", _("Removed")),
]
DEVICE_CATEGORY_CHOICES = [
    ("bike", _("Bike")),
    ("scooter", _("Scooter")),
    ("car", _("Car")),
]
DEVICE_PROPULSION_CHOICES = [
    ("electric", _("Electric")),
    ("combustion", _("Combustion")),
]

EVENT_TYPE_CHOICES = [
    ("service_start", _("Service start")),
    ("trip_end", _("Trip end")),
    ("rebalance_drop_off", _("Rebalance drop_off")),
    ("maintenance_drop_off", _("Maintenance drop off")),
    ("cancel_reservation", _("Cancel reservation")),
    ("reserve", _("Reserve")),
    ("trip start", _("Trip start")),
    ("trip enter", _("Trip enter")),
    ("trip_leave", _("Trip leave")),
    ("register", _("Register")),
    ("low_battery", _("Low battery")),
    ("maintenance", _("Maintenance")),
    ("service_end", _("Service end")),
    ("rebalance_pick_up", _("Rebalance pick up")),
    ("maintenance_pick_up", _("Maintenance pick up")),
    ("deregister", _("Deregister")),
    # this last event is not in the MDS spec
    ("telemetry", _("Received telemetry")),
]

EVENT_INGESTION_SOURCES = [
    ("push", "push"),  # pushed by the provider
    ("pull", "pull"),  # pulled by agency (e.g. from the provider's API
]
